@@ -1,4 +1,4 @@
-# Copyright 2018 Google LLC
+# Copyright 2018 The JAX Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -21,7 +21,7 @@ import itertools
 import os
 import sys
 import threading
-from typing import Any, List, Callable, NamedTuple, Iterator, Optional
+from typing import Any, List, Callable, Hashable, NamedTuple, Iterator, Optional
 import warnings
 
 from absl import logging
@@ -76,8 +76,6 @@ class Config:
     self._contextmanager_flags = set()
     self._update_hooks = {}
 
-    self.omnistaging_enabled = True  # TODO(mattjj): remove this
-
   def update(self, name, val):
     if self.use_absl:
       setattr(self.absl_flags.FLAGS, name, val)
@@ -172,22 +170,6 @@ class Config:
       self.complete_absl_config(absl.flags)
       already_configured_with_absl = True
 
-      if not FLAGS.jax_omnistaging:
-        raise Exception(
-          "Disabling of omnistaging is no longer supported in JAX version 0.2.12 and higher: "
-          "see https://github.com/google/jax/blob/main/docs/design_notes/omnistaging.md.\n"
-          "To remove this warning, unset the JAX_OMNISTAGING environment variable.")
-
-  def enable_omnistaging(self):
-    warnings.warn(
-        "enable_omnistaging() is a no-op in JAX versions 0.2.12 and higher;\n"
-        "see https://github.com/google/jax/blob/main/docs/design_notes/omnistaging.md")
-
-  def disable_omnistaging(self):
-    raise Exception(
-      "Disabling of omnistaging is no longer supported in JAX version 0.2.12 and higher: "
-      "see https://github.com/google/jax/blob/main/docs/design_notes/omnistaging.md.")
-
   def define_bool_state(
     self, name: str, default: bool, help: str, *,
     update_global_hook: Optional[Callable[[bool], None]] = None,
@@ -392,9 +374,15 @@ class Config:
 
     Values included in this set should also most likely be included in
     the C++ JIT state, which is handled separately."""
-    return (self.x64_enabled, self.jax_numpy_rank_promotion,
+    tls = jax_jit.thread_local_state()
+    axis_env_state = ()
+    context = tls.extra_jit_context
+    if context and context.axis_env_state is not None:
+      axis_env_state = context.axis_env_state
+    return (axis_env_state, self.x64_enabled, self.jax_numpy_rank_promotion,
             self.jax_default_matmul_precision, self.jax_dynamic_shapes,
-            self.jax_numpy_dtype_promotion, self.jax_default_device)
+            self.jax_numpy_dtype_promotion, self.jax_default_device,
+            self.jax_array)
 
 class NoDefault: pass
 no_default = NoDefault()
@@ -501,6 +489,7 @@ class _ThreadLocalExtraJitContext(NamedTuple):
   `_update_thread_local_jit_state` in core.py to prevent circular imports.
   """
   dynamic_trace_state: Optional[Any] = None
+  axis_env_state: Hashable = ()
   numpy_rank_promotion: Optional[str] = None
   numpy_dtype_promotion: Optional[str] = None
   default_matmul_precision: Optional[Any] = None
@@ -533,14 +522,6 @@ def update_thread_local_jit_state(**kw):
   tls.extra_jit_context = _thread_local_state_cache.canonicalize(tmp)
 
 
-# TODO(mattjj): remove all uses of this flag
-flags.DEFINE_bool(
-    'jax_omnistaging',
-    bool_env('JAX_OMNISTAGING', True),
-    help=('Deprecated. Setting this flag to False raises an error. Setting it '
-          'to True has no effect.'),
-)
-
 flags.DEFINE_integer(
     'jax_tracer_error_num_traceback_frames',
     int_env('JAX_TRACER_ERROR_NUM_TRACEBACK_FRAMES', 5),
@@ -602,6 +583,15 @@ jax2tf_associative_scan_reductions = config.define_bool_state(
     )
 )
 
+jax2tf_default_experimental_native_lowering = config.define_bool_state(
+    name='jax2tf_default_experimental_native_lowering',
+    default=bool_env('JAX2TF_DEFAULT_EXPERIMENTAL_NATIVE_LOWERING', False),
+    help=(
+        'DO NOT USE, highly experimental. Sets the default value of the '
+        'experimental_native_lowering parameter to jax2tf.convert.'
+    )
+)
+
 jax_platforms = config.define_string_state(
     name='jax_platforms',
     default=None,
@@ -662,9 +652,20 @@ parallel_functions_output_gda = config.define_bool_state(
     default=False,
     help='If True, pjit will output GDAs.')
 
+def _update_jax_array_global(val):
+  if lib.xla_extension_version >= 92:
+    lib.jax_jit.global_state().jax_array = val
+
+def _update_jax_array_thread_local(val):
+  if lib.xla_extension_version >= 92:
+    lib.jax_jit.thread_local_state().jax_array = val
+
 jax_array = config.define_bool_state(
     name='jax_array',
     default=False,
+    upgrade=True,
+    update_global_hook = _update_jax_array_global,
+    update_thread_local_hook = _update_jax_array_thread_local,
     help=('If True, new pjit behavior will be enabled and `jax.Array` will be '
           'used.'))
 
@@ -860,7 +861,7 @@ bcoo_cusparse_lowering = config.define_bool_state(
 # if the intended backend can handle lowering the result
 config.define_bool_state(
     name='jax_dynamic_shapes',
-    default=False,
+    default=bool(os.getenv('JAX_DYNAMIC_SHAPES', '')),
     help=('Enables experimental features for staging out computations with '
           'dynamic shapes.'),
     update_global_hook=lambda val: \
@@ -873,6 +874,12 @@ config.define_bool_state(
     default=True,
     help='Enable using the context manager-based name stack.')
 
+config.define_bool_state(
+    name='jax_unique_mhlo_module_names',
+    default=False,
+    help='Enables the generation of unique MHLO module names. This is useful '
+    'to clients that expect modules to have unique names (e.g, trace data).')
+
 # This flag is temporary during rollout of the remat barrier.
 # TODO(parkers): Remove if there are no complaints.
 config.define_bool_state(
@@ -880,23 +887,38 @@ config.define_bool_state(
     default=(lib.version >= (0, 3, 6)),
     help=('Enables using optimization-barrier op for lowering remat.'))
 
-# TODO(mattjj): remove after May 19 2022, NeurIPS submission deadline
-config.define_bool_state(
-    name='after_neurips',
-    default=False,
-    upgrade=True,
-    help='Gate changes until after NeurIPS 2022 deadline.')
-
 # TODO(b/205307544): Remove flag once coordination service has rolled out.
 config.define_bool_state(
     name='jax_coordination_service',
-    default=False,
+    default=(lib.xla_extension_version >= 80),
     help=(
          'Use coordination service (experimental) instead of the default PjRT '
          'distributed runtime.'
     )
 )
 
+config.define_bool_state(
+    name='jax_experimental_subjaxpr_lowering_cache',
+    default=False,
+    help='Enable using a cache for lowering subjaxprs.')
+
+# TODO(sharadmv,mattjj): set default to True, then remove
+config.define_bool_state(
+    name='jax_eager_pmap',
+    default=True,
+    upgrade=True,
+    help='Enable eager-mode pmap when jax_disable_jit is activated.')
+
+config.define_bool_state(
+    name='jax_experimental_unsafe_xla_runtime_errors',
+    default=False,
+    help=('Enable XLA runtime errors for jax.experimental.checkify.checks '
+          'on CPU and GPU. These errors are async, might get lost and are not '
+          'very readable. But, they crash the computation and enable you '
+          'to write jittable checks without needing to checkify. Does not '
+          'work under pmap/pjit.')
+)
+
 @contextlib.contextmanager
 def explicit_device_put_scope() -> Iterator[None]:
   """Indicates that the current context is an explicit device_put*() call."""
