@@ -1,4 +1,4 @@
-# Copyright 2018 Google LLC
+# Copyright 2018 The JAX Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -32,15 +32,16 @@ traceback_util.register_exclusion(__file__)
 T = TypeVar("T")
 U = TypeVar("U")
 
-if TYPE_CHECKING:
-  PyTreeDef = pytree.PyTreeDef
-else:
-  PyTreeDef = Any
+PyTreeDef = pytree.PyTreeDef
 
 
 def tree_flatten(tree, is_leaf: Optional[Callable[[Any], bool]] = None):
   """Flattens a pytree.
 
+  The flattening order (i.e. the order of elements in the output list)
+  is deterministic, corresponding to a left-to-right depth-first tree
+  traversal.
+
   Args:
     tree: a pytree to flatten.
     is_leaf: an optionally specified function that will be called at each
@@ -88,11 +89,14 @@ def treedef_children(treedef):
 def treedef_is_leaf(treedef):
   return treedef.num_nodes == 1
 
+def treedef_is_strict_leaf(treedef):
+  return treedef.num_nodes == 1 and treedef.num_leaves == 1
+
 def all_leaves(iterable, is_leaf: Optional[Callable[[Any], bool]] = None):
   """Tests whether all elements in the given iterable are all leaves.
 
   >>> tree = {"a": [1, 2, 3]}
-  >>> assert all_leaves(jax.tree_leaves(tree))
+  >>> assert all_leaves(jax.tree_util.tree_leaves(tree))
   >>> assert not all_leaves([tree])
 
   This function is useful in advanced cases, for example if a library allows
@@ -120,7 +124,7 @@ def register_pytree_node(nodetype: Type[T],
                          unflatten_func: Callable[[_AuxData, _Children], T]):
   """Extends the set of types that are considered internal nodes in pytrees.
 
-  See `example usage <pytrees.html>`_.
+  See :ref:`example usage <pytrees>`.
 
   Args:
     nodetype: a Python type to treat as an internal pytree node.
@@ -195,12 +199,6 @@ def tree_map(f: Callable[..., Any], tree: Any, *rest: Any,
   all_leaves = [leaves] + [treedef.flatten_up_to(r) for r in rest]
   return treedef.unflatten(f(*xs) for xs in zip(*all_leaves))
 
-def tree_multimap(*args, **kwargs):
-  """Deprecated alias of :func:`jax.tree_util.tree_map`"""
-  warnings.warn('jax.tree_util.tree_multimap() is deprecated. Please use jax.tree_util.tree_map() '
-                'instead as a drop-in replacement.', FutureWarning)
-  return tree_map(*args, **kwargs)
-
 def build_tree(treedef, xs):
   return treedef.from_iterable_tree(xs)
 
@@ -431,7 +429,7 @@ class FlattenedKeyPathEntry(KeyPathEntry):  # fallback
     return f'[<flat index {self.key}>]'
 
 def _child_keys(pytree: Any) -> List[KeyPathEntry]:
-  assert not treedef_is_leaf(tree_structure(pytree))
+  assert not treedef_is_strict_leaf(tree_structure(pytree))
   handler = _keypath_registry.get(type(pytree))
   if handler:
     return handler(pytree)
@@ -459,7 +457,7 @@ def _prefix_error(key_path: KeyPath, prefix_tree: Any, full_tree: Any,
                   is_leaf: Optional[Callable[[Any], bool]] = None,
                   ) -> Iterable[Callable[[str], ValueError]]:
   # A leaf is a valid prefix of any tree:
-  if treedef_is_leaf(tree_structure(prefix_tree, is_leaf=is_leaf)): return
+  if treedef_is_strict_leaf(tree_structure(prefix_tree, is_leaf=is_leaf)): return
 
   # The subtrees may disagree because their roots are of different types:
   if type(prefix_tree) != type(full_tree):
@@ -518,3 +516,22 @@ def _prefix_error(key_path: KeyPath, prefix_tree: Any, full_tree: Any,
     f"equal pytree nodes gave differing keys: {keys} and {keys_}"
   for k, t1, t2 in zip(keys, prefix_tree_children, full_tree_children):
     yield from _prefix_error(key_path + k, t1, t2)
+
+
+# TODO(jakevdp) remove these deprecated wrappers & their imports in jax/__init__.py
+def _deprecate(f):
+  @functools.wraps(f)
+  def wrapped(*args, **kwargs):
+    warnings.warn(f"jax.{f.__name__} is deprecated, and will be removed in a future release. "
+                  f"Use jax.tree_util.{f.__name__} instead.",
+                  category=FutureWarning, stacklevel=2)
+    return f(*args, **kwargs)
+  return wrapped
+
+def __getattr__(name):
+  prefix = "_deprecated_"
+  if name.startswith(prefix):
+    name = name[len(prefix):]
+    return _deprecate(globals()[name])
+  else:
+    raise AttributeError(f"module {__name__} has no attribute {name!r}")
