@@ -1,4 +1,4 @@
-# Copyright 2021 Google LLC
+# Copyright 2021 The JAX Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -15,26 +15,48 @@
 
 import abc
 import asyncio
+import itertools
+from functools import partial
 import re
 import threading
-import time
-from typing import Callable
+from typing import Callable, Sequence, Optional
 from absl import logging
 
 import jax
 from jax._src import distributed
-from jax._src.util import prod
+from jax._src.config import config
 from jax.experimental import global_device_array as gda
+from jax.experimental import array
+from jax.experimental import sharding
 from jax.experimental.maps import Mesh
 import jax.numpy as jnp
 import numpy as np
 import tensorstore as ts
-from etils import epath
 
 
 TS_CONTEXT = ts.Context({'file_io_concurrency': {'limit': 128}})
 _REMOVED_VALUE = 'Value removed'
 _CHECKPOINT_SUCCESS = 'checkpoint_write_success'
+_module_unique_count = itertools.count()
+
+
+async def create_async_array_from_callback(
+    global_shape: array.Shape,
+    inp_sharding: sharding.XLACompatibleSharding,
+    data_callback: Callable[[array.Index], asyncio.Future],
+):
+  device_to_index_map = inp_sharding.devices_indices_map(global_shape)
+  addressable_da = inp_sharding._addressable_device_assignment
+  future_arrays = [data_callback(device_to_index_map[d])  # type: ignore
+                   for d in addressable_da]
+  # Pause here and come back to `from_async_callback()` when future_arrays are
+  # ready. device_put cannot happen with future_arrays.
+  local_arrays = await asyncio.gather(*future_arrays)
+
+  dbs = [jax.device_put(array, device)
+         for array, device in zip(local_arrays, addressable_da)]
+  aval = jax.ShapedArray(global_shape, dbs[0].dtype)
+  return array.Array(aval, inp_sharding, dbs, committed=True)
 
 
 async def create_async_gda_from_callback(
@@ -58,19 +80,22 @@ async def create_async_gda_from_callback(
                                gda._GdaFastPathArgs(global_idx_rid, local_devices))
 
 
-def _get_metadata(gda):
-  if gda.dtype == jnp.bfloat16:
+def _get_metadata(arr):
+  if arr.dtype == jnp.bfloat16:
     # Tensorstore uses 'bfloat16', not '<V2'.
     dtype = 'bfloat16'
   else:
-    dtype = np.dtype(gda.dtype).str
-
+    dtype = np.dtype(arr.dtype).str
+  if isinstance(arr, array.Array):
+    local_shape = arr._arrays[0].shape
+  else:
+    local_shape = arr.local_data(0).shape
   return {
       'compressor': {
           'id': 'gzip'
       },
-      'shape': gda.shape,
-      'chunks': np.array(np.maximum(1, gda.local_data(0).shape)),
+      'shape': arr.shape,
+      'chunks': np.array(np.maximum(1, local_shape)),
       'dtype': dtype,
   }
 
@@ -99,12 +124,40 @@ def get_tensorstore_spec(ckpt_path: str):
   return spec
 
 
-async def async_serialize(gda_inp: gda.GlobalDeviceArray, tensorstore_spec,
-                          commit_future=None):
+# Lifted from T5X.
+class _LimitInFlightBytes:
+  """Limits in-flight bytes when reading/writing checkpoints per process."""
+
+  def __init__(self, num_bytes):
+    self._max_bytes = num_bytes
+    self._available_bytes = num_bytes
+    self._cv = asyncio.Condition(lock=asyncio.Lock())
+
+  async def wait_for_bytes(self, requested_bytes):
+    if requested_bytes >= self._max_bytes:
+      raise ValueError('Requested more bytes than we reserved space for: '
+                       f'{requested_bytes} > {self._max_bytes}')
+    async with self._cv:
+      await self._cv.wait_for(lambda: self._available_bytes > requested_bytes)
+      self._available_bytes -= requested_bytes
+      assert self._available_bytes >= 0
+
+  async def release_bytes(self, requested_bytes):
+    async with self._cv:
+      self._available_bytes += requested_bytes
+      assert self._available_bytes <= self._max_bytes
+      self._cv.notify_all()
+
+
+async def async_serialize(arr_inp, tensorstore_spec, commit_future=None):
+  if (isinstance(arr_inp, array.Array) and jax.process_count() > 1 and
+      arr_inp.is_fully_addressable()):
+    raise ValueError('Passing fully addressable Arrays to a multi-host '
+                     'serialization is not allowed.')
   # 'metadata' may not be present at the top level (for example, if we are using
   # a 'cast' driver).
   if not _spec_has_metadata(tensorstore_spec):
-    tensorstore_spec['metadata'] = _get_metadata(gda_inp)
+    tensorstore_spec['metadata'] = _get_metadata(arr_inp)
 
   if jax.process_index() == 0:
     open_future = ts.open(
@@ -134,61 +187,110 @@ async def async_serialize(gda_inp: gda.GlobalDeviceArray, tensorstore_spec,
       else:
         await write_future.commit
 
-  future_write_state = jax.tree_util.tree_map(_write_array,
-                                              gda_inp.local_shards)
+  if isinstance(arr_inp, array.Array):
+    local_shards = arr_inp.addressable_shards
+  else:
+    local_shards = arr_inp.local_shards
+  future_write_state = jax.tree_util.tree_map(_write_array, local_shards)
   return await asyncio.gather(*future_write_state)
 
 
-def run_serialization(gdas, tensorstore_specs):
+def run_serialization(arrays, tensorstore_specs):
   async def _run_serializer():
-    future_writer = jax.tree_map(async_serialize, gdas, tensorstore_specs)
+    future_writer = jax.tree_util.tree_map(async_serialize, arrays, tensorstore_specs)
     return await asyncio.gather(*future_writer)
   asyncio.run(_run_serializer())
 
 
-async def async_deserialize(mesh, mesh_axes, tensorstore_spec,
-                            global_shape=None, dtype=None):
-  t = ts.open(ts.Spec(tensorstore_spec), open=True, context=TS_CONTEXT).result()
-  shape = t.shape if global_shape is None else global_shape
-  requires_padding = prod(shape) > prod(t.shape)
+def estimate_read_memory_footprint(t: ts.TensorStore) -> int:
+  rank = t.rank
+  num_bytes = t.dtype.numpy_dtype.itemsize
+  chunk_template = t.chunk_layout.read_chunk_template
+  origin = t.domain.origin
+  shape = t.domain.shape
+  chunk_origin = chunk_template.origin
+  chunk_shape = chunk_template.shape
 
-  if requires_padding:
-    new_shard_shape = gda.get_shard_shape(tuple(shape), mesh, mesh_axes)
+  # Some TensorStore drivers are not chunked, e.g. the inline 'array' driver.
+  # For those, instead of returning a near-infinite memory footprint, estimate
+  # the footprint as the entire shape.
+  for i in range(rank):
+    if not chunk_template[i].finite:
+      return t.domain.size * num_bytes
+
+  # Otherwise, if we have a chunked driver, estimate based on chunk size.
+  for i in range(rank):
+    origin_value = origin[i]
+    chunk_origin_value = chunk_origin[i]
+    chunk_size = chunk_shape[i]
+    lower = origin_value - chunk_origin_value
+    upper = origin_value + shape[i] - chunk_origin_value
+    lower_aligned = lower // chunk_size * chunk_size
+    upper_aligned = -(-upper // chunk_size) * chunk_size
+    num_bytes *= (upper_aligned - lower_aligned)
+
+  return num_bytes
+
+
+async def async_deserialize(mesh, mesh_axes, tensorstore_spec,
+                            global_shape=None, dtype=None,
+                            byte_limiter: Optional[_LimitInFlightBytes] = None):
+  t = await ts.open(ts.Spec(tensorstore_spec), open=True, context=TS_CONTEXT)
+  shape = t.shape if global_shape is None else global_shape
+  new_shard_shape = gda.get_shard_shape(tuple(shape), mesh, mesh_axes)
 
   async def cb(index):
-    if requires_padding:
-      # This is needed because the shape the array was saved with is smaller
-      # than the requested shape of the array in which it will be reloaded. So
-      # the extra values will be filled with 0s.
-      out = np.zeros(new_shard_shape, dtype=t.dtype.numpy_dtype)
-      requested_domain = ts.IndexTransform(input_shape=shape)[index].domain
-      restricted_domain = t.domain.intersect(requested_domain)
-      await ts.array(out)[ts.d[:].translate_to[requested_domain.origin]][restricted_domain].write(t[restricted_domain])
-    else:
-      out = await t[index].read()
+    # This maybe needed because the shape the array was saved with is smaller
+    # than the requested shape of the array in which it will be reloaded. So
+    # the extra values will be filled with 0s.
+    out = np.zeros(new_shard_shape, dtype=t.dtype.numpy_dtype)
+    requested_domain = ts.IndexTransform(input_shape=shape)[index].domain
+    restricted_domain = t.domain.intersect(requested_domain)
+
+    requested_bytes = estimate_read_memory_footprint(t[restricted_domain])
+
+    # Limit the bytes read for every shard.
+    if byte_limiter is not None:
+      await byte_limiter.wait_for_bytes(requested_bytes)
+
+    await ts.array(out)[ts.d[:].translate_to[requested_domain.origin]][restricted_domain].write(
+        t[restricted_domain])
 
     if dtype is not None:
       # Cast while reloading on process to avoid 2 copies on device if the
       # casting is done on device.
-      return out.astype(dtype)
+      out = out.astype(dtype)
+
+    if byte_limiter is not None:
+      await byte_limiter.release_bytes(requested_bytes)
     return out
 
-  return await create_async_gda_from_callback(tuple(shape), mesh, mesh_axes, cb)
+  if config.jax_array:
+    inp_sharding = sharding.MeshPspecSharding(mesh, mesh_axes)
+    return await create_async_array_from_callback(tuple(shape), inp_sharding, cb)
+  else:
+    return await create_async_gda_from_callback(tuple(shape), mesh, mesh_axes, cb)
 
 
 def run_deserialization(global_meshes, mesh_axes, tensorstore_specs,
-                        global_shapes=None, dtypes=None):
+                        global_shapes=None, dtypes=None, concurrent_gb=32):
+  concurrent_bytes = concurrent_gb * 10**9
+
   async def _run_deserializer():
-    future_gdas = jax.tree_map(
-        async_deserialize, global_meshes, mesh_axes, tensorstore_specs,
+    # Object should be created once per process.
+    byte_limiter = _LimitInFlightBytes(concurrent_bytes)
+
+    future_arrays = jax.tree_util.tree_map(
+        partial(async_deserialize, byte_limiter=byte_limiter),
+        global_meshes, mesh_axes, tensorstore_specs,
         [None] * len(tensorstore_specs) if global_shapes is None else global_shapes,
         [None] * len(tensorstore_specs) if dtypes is None else dtypes)
-    return await asyncio.gather(*future_gdas)
+    return await asyncio.gather(*future_arrays)
   return asyncio.run(_run_deserializer())
 
 
 def _get_key(key: str):
-  return f'checkpoint_{key}'
+  return f'tensorstore_checkpoint_{key}'
 
 
 class GlobalAsyncCheckpointManagerBase(metaclass=abc.ABCMeta):
@@ -244,11 +346,8 @@ class GlobalAsyncCheckpointManagerBase(metaclass=abc.ABCMeta):
     """Blocks until serialization has finished."""
 
   @abc.abstractmethod
-  # TODO(b/233793426): Try removing temp_checkpoint_dir and final_checkpoint_dir
-  # from the API and use a callback instead. This will affect how async
-  # mechanism works.
-  def serialize(self, gdas, tensorstore_specs, *, temp_checkpoint_dir,
-                final_checkpoint_dir):
+  def serialize(self, arrays, tensorstore_specs, *,
+                on_commit_callback: Callable[[], None]):
     """Serializes GDAs to TensorStore."""
 
   @abc.abstractmethod
@@ -257,8 +356,7 @@ class GlobalAsyncCheckpointManagerBase(metaclass=abc.ABCMeta):
     """Deserializes GDAs from TensorStore."""
 
 
-class GlobalAsyncCheckpointManager(GlobalAsyncCheckpointManagerBase):
-  """Responsible for serializing GDAs via TensorStore."""
+class AsyncManager:
 
   def __init__(self, timeout_secs=300):
     self._timeout_secs = timeout_secs
@@ -273,43 +371,46 @@ class GlobalAsyncCheckpointManager(GlobalAsyncCheckpointManagerBase):
                        '`jax.distributed.initialize()` at the start of your '
                        'program.')
     self._client = distributed.global_state.client
-    self._final_ckpt_dir = None
+    self._count = None
 
   def __del__(self):
     if self._thread is not None and self._thread.is_alive():
       logging.warning('Please add `.wait_until_finished()` in the main thread '
                       'before your program finishes because there is a '
                       'possibility of losing errors raised if the '
-                      'GlobalAsyncCheckpointManager is deleted before '
-                      'serialization is completed.')
+                      'this class is deleted before writing is completed.')
 
-  def _thread_func(self, temp_checkpoint_dir, final_checkpoint_dir):
+  def _thread_func(self):
     try:
-      for future in self._commit_futures:
-        for f in future:
-          f.result()
-
       current_process = jax.process_index()
-      logging.info('Commit to storage layer has completed by process: %s',
+      logging.info('Starting commit to storage layer by process: %s',
+                   current_process)
+      for future in self._commit_futures:
+        future.result()
+      logging.info('Finished committing to storage layer by process: %s',
                    current_process)
 
       # All processes will wait at the barrier. When all processes are at the
       # barrier, the barrier will be satisfied. If not, then it will timeout.
-      self._client.wait_at_barrier(self._final_ckpt_dir, self._timeout_in_ms)
-      logging.info('Finished waiting at barrier for process %s', current_process)
+      key_for_barrier = _get_key(self._count)
+      logging.info('Key used for barrier is %s for process %s',
+                   key_for_barrier, current_process)
+      self._client.wait_at_barrier(key_for_barrier, self._timeout_in_ms)
+      logging.info('Finished waiting at barrier for process %s',
+                   current_process)
 
       if current_process == 0:
-        logging.info('Renaming %s to %s', temp_checkpoint_dir, final_checkpoint_dir)
-        epath.Path(temp_checkpoint_dir).rename(final_checkpoint_dir)
-        logging.info('Finished saving GDA checkpoint to `%s`.', final_checkpoint_dir)
-        self._client.key_value_set(_get_key(self._final_ckpt_dir), _CHECKPOINT_SUCCESS)
+        self._on_commit_callback()
+        self._client.key_value_set(key_for_barrier, _CHECKPOINT_SUCCESS)
+
     except Exception as e:
       self._exception = e
 
-  def _start_async_commit(self, temp_checkpoint_dir, final_checkpoint_dir):
-    self._thread = threading.Thread(
-        target=self._thread_func,
-        args=(temp_checkpoint_dir, final_checkpoint_dir))
+  def _start_async_commit(self, on_commit_callback):
+    self._count = next(_module_unique_count)
+
+    self._on_commit_callback = on_commit_callback
+    self._thread = threading.Thread(target=self._thread_func)
     self._thread.start()
 
   def check_for_errors(self):
@@ -326,15 +427,21 @@ class GlobalAsyncCheckpointManager(GlobalAsyncCheckpointManagerBase):
 
     self.check_for_errors()
 
-    if self._final_ckpt_dir is not None:
+    if self._count is not None:
       # Block until process 0 writes success value to the key value store.
       # If it fails to write it, then `blocking_key_value_get` will time out.
-      self._client.blocking_key_value_get(_get_key(self._final_ckpt_dir),
-                                          self._timeout_in_ms)
+      self._client.blocking_key_value_get(
+          _get_key(self._count), self._timeout_in_ms)
 
-  def serialize(self, gdas, tensorstore_specs, *, temp_checkpoint_dir,
-                final_checkpoint_dir):
-    """Serializes GlobalDeviceArrays via TensorStore asynchronously.
+  def _add_futures(self, futures: Sequence[asyncio.Future]):
+    self._commit_futures = futures
+
+
+class GlobalAsyncCheckpointManager(AsyncManager, GlobalAsyncCheckpointManagerBase):
+  """Responsible for serializing GDAs via TensorStore."""
+
+  def serialize(self, arrays, tensorstore_specs, *, on_commit_callback):
+    """Serializes GlobalDeviceArrays or Arrays via TensorStore asynchronously.
 
     TensorStore writes to a storage layer in 2 steps:
     *  Reading/copying from the source after which the source can be modified.
@@ -346,8 +453,9 @@ class GlobalAsyncCheckpointManager(GlobalAsyncCheckpointManagerBase):
     finish in a separate thread allowing other computation to proceed.
 
     Args:
-      gdas: GlobalDeviceArrays that should be serialized.
-      tensorstore_specs: TensorStore specs that are used to serialize GDAs.
+      arrays: GlobalDeviceArrays or Arrays that should be serialized.
+      tensorstore_specs: TensorStore specs that are used to serialize GDAs or
+        Arrays.
       temp_checkpoint_dir: Temporary checkpoint directory where the checkpoints
         will be written.
       final_checkpoint_dir: Final checkpoint directory where the checkpoints
@@ -356,20 +464,23 @@ class GlobalAsyncCheckpointManager(GlobalAsyncCheckpointManagerBase):
     logging.info('Waiting for previous serialization to finish.')
     self.wait_until_finished()
 
-    self._commit_futures = [[] for _ in range(len(tensorstore_specs))]
+    commit_futures = [[] for _ in range(len(tensorstore_specs))]
 
     async def _run_serializer():
-      future_writer = jax.tree_map(async_serialize, gdas,
-                                   tensorstore_specs, self._commit_futures)
+      future_writer = jax.tree_util.tree_map(
+          async_serialize, arrays, tensorstore_specs, commit_futures)
       return await asyncio.gather(*future_writer)
+
     asyncio.run(_run_serializer())
 
+    self._add_futures(jax.tree_util.tree_flatten(commit_futures)[0])
+
     # Used in wait_until_finished to check on process != 0, if the checkpoint
     # has finished writing.
-    self._final_ckpt_dir = final_checkpoint_dir
-    self._start_async_commit(temp_checkpoint_dir, final_checkpoint_dir)
+    self._start_async_commit(on_commit_callback)
 
   def deserialize(self, global_meshes, mesh_axes, tensorstore_specs,
                   global_shapes=None, dtypes=None):
+    self.wait_until_finished()
     return run_deserialization(global_meshes, mesh_axes, tensorstore_specs,
                                global_shapes, dtypes)
