@@ -1,4 +1,4 @@
-# Copyright 2020 Google LLC
+# Copyright 2020 The JAX Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,28 +14,32 @@
 """Tests for JAX2TF converted.
 
 Specific JAX primitive conversion tests are in primitives_test."""
-
+import collections
+import os
+from typing import Callable, Dict, Optional, Tuple
 import unittest
-from typing import Dict, Tuple
 
+from absl import logging
 from absl.testing import absltest
 from absl.testing import parameterized
 
-from collections import OrderedDict
-import os
-
 import jax
 from jax import ad_checkpoint
 from jax import dtypes
 from jax import lax
 from jax import numpy as jnp
+from jax._src import lib as jaxlib
+from jax._src import source_info_util
 from jax._src import test_util as jtu
+import jax._src.lib.xla_bridge
 from jax.config import config
 from jax.experimental import jax2tf
+from jax.experimental.global_device_array import GlobalDeviceArray
 from jax.experimental.jax2tf.tests import tf_test_util
-from jax._src import source_info_util
-import jax._src.lib.xla_bridge
-
+from jax.experimental.pjit import FROM_GDA
+from jax.experimental.pjit import pjit
+from jax.interpreters import mlir
+from jax.interpreters.pxla import PartitionSpec as P
 import numpy as np
 import tensorflow as tf  # type: ignore[import]
 # pylint: disable=g-direct-tensorflow-import
@@ -47,9 +51,13 @@ config.parse_flags_with_absl()
 
 class Jax2TfTest(tf_test_util.JaxToTfTestCase):
 
+  def test_empty(self):
+    f_jax = lambda x, y: x
+    self.ConvertAndCompare(f_jax, 0.7, 1)
+
   def test_basics(self):
     f_jax = lambda x: jnp.sin(jnp.cos(x))
-    _, res_tf = self.ConvertAndCompare(f_jax, 0.7)
+    self.ConvertAndCompare(f_jax, 0.7)
 
   def test_input_output_naming(self):
     @jax2tf.convert
@@ -363,7 +371,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     default_float_type = jax2tf.dtype_of_val(4.)
     x = tf.Variable([4.], dtype=default_float_type)
     y = tf.Variable([4., 5.], dtype=default_float_type)
-    inputs = OrderedDict()
+    inputs = collections.OrderedDict()
     inputs['r'] = x
     inputs['d'] = y
     with tf.GradientTape(persistent=True) as tape:
@@ -480,8 +488,8 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
       dict(testcase_name=f"function={with_function}",
            with_function=with_function)
       for with_function in [False, True]))
-  def test_gradients_unused_argument_readme(self, with_function=True):
-    # x2 and x3 are not used. x3 has integer type.
+  def test_gradients_unused_argument_readme(self, with_function=False):
+    # x1 and x3 are not used. x3 has integer type.
     def fn(x0, x1, x2, x3):
       return x0 * 0. + x2 * 2.
 
@@ -528,7 +536,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
       dict(testcase_name=f"function={with_function}",
            with_function=with_function)
       for with_function in [False, True]))
-  def test_gradients_int_argument(self, with_function=True):
+  def test_gradients_int_argument(self, with_function=False):
     # https://github.com/google/jax/issues/6975
     # Also issue #6975.
     # An expanded version of test_gradients_unused_argument
@@ -690,9 +698,9 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     self.assertNotEqual(type(m.a), list)
     self.assertNotEqual(type(m.b), tuple)
     self.assertNotEqual(type(m.c), dict)
-    self.assertLen(jax.tree_leaves(m.a), 2)
-    self.assertLen(jax.tree_leaves(m.b), 2)
-    self.assertLen(jax.tree_leaves(m.c), 2)
+    self.assertLen(jax.tree_util.tree_leaves(m.a), 2)
+    self.assertLen(jax.tree_util.tree_leaves(m.b), 2)
+    self.assertLen(jax.tree_util.tree_leaves(m.c), 2)
 
   def test_issue_10586(self):
 
@@ -759,16 +767,13 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     self.TransformConvertAndCompare(f, arg, "grad")
     self.TransformConvertAndCompare(f, arg, "grad_vmap")
 
-  @parameterized.named_parameters(jtu.cases_from_list(
-    dict(testcase_name=f"_{flavor}", flavor=flavor)
-    for flavor in ["old", "new"]))
-  def test_remat(self, flavor="old"):
+  def test_remat(self):
     def f(x1):
       x2 = jnp.sin(x1)
       x3 = jnp.sin(x2)
       x4 = jnp.sin(x3)
       return x4
-    remat_f = jax.remat(f) if flavor == "old" else ad_checkpoint.checkpoint(f)
+    remat_f = ad_checkpoint.checkpoint(f)
 
     # The computation of grad_f computes "sin" 5 times, 3 for the forward pass
     # and then to rematerialize "x2" and "x3" in the backward pass.
@@ -776,17 +781,19 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     # Check that we have a Sin under a conditional
     f_tf = tf.function(jax2tf.convert(jax.grad(remat_f)), autograph=False)
     f_tf_graph = f_tf.get_concrete_function(arg).graph.as_graph_def()
-    if flavor == "old":
-      raise unittest.SkipTest("TODO: CSE widget not yet implemented for old-style remat")
     if jax.config.jax_remat_opt_barrier:
-      self.assertRegex(
-          str(f_tf_graph), r"remat_checkpoint_/XlaOptimizationBarrier")
+      if config.jax2tf_default_experimental_native_lowering:
+        self.assertRegex(
+          str(f_tf_graph), r"mhlo.optimization_barrier")
+      else:
+        self.assertRegex(
+            str(f_tf_graph), r"XlaOptimizationBarrier")
     elif config.jax_experimental_name_stack:
       self.assertRegex(str(f_tf_graph),
-                       r'transpose/jax2tf_f_/jvp/checkpoint/remat_checkpoint_/cond/branch_1_fun/Sin')
+                       r'transpose/jax2tf_f_/jvp/checkpoint/cond/branch_1_fun/Sin')
     else:
       self.assertRegex(str(f_tf_graph),
-                       r'remat_checkpoint_/switch_case/indexed_case/Sin')
+                       r'switch_case/indexed_case/Sin')
 
   def test_remat_free_var(self):
     def f(x):
@@ -813,7 +820,10 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
       return jnp.sin(1.)
     f_tf = tf.function(jax2tf.convert(f_jax), autograph=False)
     f_tf_graph = f_tf.get_concrete_function().graph.as_graph_def()
-    self.assertIn('op: "Sin"', str(f_tf_graph))
+    if config.jax2tf_default_experimental_native_lowering:
+      self.assertIn("mhlo.sine", str(f_tf_graph))
+    else:
+      self.assertIn('op: "Sin"', str(f_tf_graph))
 
   def test_convert_of_nested_independent_jit(self):
     def func(x):
@@ -885,14 +895,13 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
         return my_test_function(jnp.sin(x))
 
       out = jax2tf.convert(caller, with_gradient=False)(2.)
-      # When we use `with_gradient=False` the raw output of `caller` is passed
-      # through a `tf.raw_ops.PreventGradient` and a `tf.identity`, clobbering
-      # the name scope of the `mul` op. We need to get the grandparent of the
-      # `out` tensor to see the name scope of the result of the `mul`.
-      grandparent_op = out.op.inputs[0].op.inputs[0]
-      self.assertIn("my_test_function", grandparent_op.name)
       return out
-    run()
+    run_graph = run.get_concrete_function().graph.as_graph_def()
+    print(str(run_graph))
+    if config.jax2tf_default_experimental_native_lowering:
+      self.assertIn("my_test_function/mul", str(run_graph))
+    else:
+      self.assertIn("my_test_function/jit_fn_/Mul", str(run_graph))
 
   def test_bfloat16_constant(self):
     # Re: https://github.com/google/jax/issues/3942
@@ -917,7 +926,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
   def test_shared_constants(self):
     # Check that the constants are shared properly in converted functions
     # See https://github.com/google/jax/issues/7992.
-    const = np.ones((16, 16))
+    const = np.random.uniform(size=256).astype(np.float32)  # A shared constant
     def f(x):
       return x + const + const + const + const
 
@@ -927,7 +936,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
   def test_shared_constants_under_cond(self):
     # Check that the constants are shared properly in converted functions
     # See https://github.com/google/jax/issues/7992.
-    const = np.arange(256, dtype=np.float32)
+    const = np.random.uniform(size=256).astype(np.float32)  # A shared constant
     x = np.ones((256,), dtype=np.float32)
     def f1(x):
       return lax.cond(x[0] >= 0., lambda x: x + const, lambda x: x * const, x) + const
@@ -939,7 +948,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
 
   def test_shared_constants_under_scan(self):
     # See https://github.com/google/jax/issues/7992.
-    const = np.arange(256, dtype=np.float32)
+    const = np.random.uniform(size=256).astype(np.float32)  # A shared constant
     xs = np.ones((8, 256), dtype=np.float32)
     def f1(xs):
       res, _ = lax.scan(lambda carry, x: (carry + x + const, None),
@@ -955,7 +964,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
 
   def test_shared_constants_under_jit(self):
     # We do not share constants under jit.
-    const = np.ones((16, 16))
+    const = np.random.uniform(size=(16, 16)).astype(np.float32)  # A shared constant
     @jax.jit
     def g_jit(x):
       return x * const
@@ -1011,6 +1020,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
                         (grad_tf[0].numpy(), grad_tf[1].numpy()))
 
 
+  @jtu.skip_on_flag("jax2tf_default_experimental_native_lowering", True)
   def test_enable_xla(self):
     # Tests that enable_xla flag is properly scoped to a conversion.
     def fun(x):
@@ -1039,6 +1049,9 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     self.ConvertAndCompare(jnp.sin, jnp.zeros((2, 3), jnp.float32))
 
   def test_randint(self):
+    if jtu.device_under_test() == "gpu" and config.jax2tf_default_experimental_native_lowering:
+      raise unittest.SkipTest("randint on GPU uses custom calls; not supported")
+
     def randint():
       return jax.random.randint(
           jax.random.PRNGKey(42), shape=(), minval=0, maxval=1)
@@ -1225,8 +1238,153 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
         include_xla_op_metadata=False
     )
 
+def get_serialized_computation(
+    f_jax: Callable,
+    *args,
+    abstracted_axes: Optional[Tuple[Dict[int, str]]] = None) -> str:
+  lowered = jax.jit(f_jax, abstracted_axes=abstracted_axes).lower(*args)
+  mhlo_module = lowered.compiler_ir(dialect='mhlo')
+  mhlo_module_text = mlir.module_to_string(mhlo_module)
+  if jaxlib.version <= (0, 3, 14):
+    mhlo_module_text = jax2tf.jax2tf._fixup_mhlo_module_text(mhlo_module_text)
+  logging.info(f'Serialized ir.Module = {mhlo_module_text}')
+  return mhlo_module_text
+
+
+class XlaCallModuleTest(tf_test_util.JaxToTfTestCase):
+  """Unit tests for XlaCallModule. Will move these eventually to TF."""
+  def test_simple(self):
+
+    def f_jax(x):
+      return jnp.sin(x)
+
+    x = np.ones((2, 3), dtype=np.float32)
+
+    jax_res = f_jax(x)
+    res = tfxla.call_module([x],
+                            module=get_serialized_computation(f_jax, x),
+                            Tout=[jax_res.dtype],
+                            Sout=[jax_res.shape])
+    self.assertAllClose(tf.nest.map_structure(lambda t: t.numpy(), res),
+                        [jax_res])
+
+  def test_while(self):
+    # With nested computation
+    def f_jax(count, x):
+      return lax.while_loop(lambda carry: carry[0] < count, lambda carry:
+                            (carry[0] + 1, carry[1] + 1.), (0, x))[1]
+
+    count = np.int32(5)
+    x = np.ones((2, 3), dtype=np.float32)
+
+    jax_res = f_jax(count, x)
+    res = tfxla.call_module([count, x],
+                            module=get_serialized_computation(f_jax, count, x),
+                            Tout=[jax_res.dtype],
+                            Sout=[jax_res.shape])
+    self.assertAllClose(tf.nest.map_structure(lambda t: t.numpy(), res),
+                        [jax_res])
+
+  def test_multiple_args_results(self):
+
+    def f_jax(x1, x2):
+      return (jnp.sin(x1), jnp.cos(x2))
+
+    x1 = np.ones((2, 3), dtype=np.float32)
+    x2 = np.ones((3, 4), dtype=np.float32)
+
+    jax_res = f_jax(x1, x2)
+
+    def f_tf(x1_tf, x2_tf):
+      return tfxla.call_module([x1_tf, x2_tf],
+                               module=get_serialized_computation(f_jax, x1, x2),
+                               Tout=[jax_res[0].dtype, jax_res[1].dtype],
+                               Sout=[jax_res[0].shape, jax_res[1].shape])
+
+    res = tf.function(f_tf, jit_compile=True, autograph=False)(x1, x2)
+    self.assertAllClose(tf.nest.map_structure(lambda t: t.numpy(), res),
+                        jax_res)
+
+  @unittest.skip("TODO(necula): 'mhlo.dynamic_iota' op can't be translated to XLA HLO")
+  def test_shape_poly_arange(self):
+    if not config.jax_dynamic_shapes:
+      raise unittest.SkipTest("jax_dynamic_shapes must be enabled")
+    def f_jax(x):  # x: f32[b]
+      return jnp.arange(x.shape[0]) + x
+
+    x1 = np.ones((5,), dtype=np.float32)
+    jax_res = f_jax(x1)
+
+    def f_tf(x1_tf):
+      return tfxla.call_module([x1_tf],
+                               module=get_serialized_computation(
+                                   f_jax, x1,
+                                   abstracted_axes=({
+                                       0: 'b'
+                                   },)),
+                               Tout=[jax_res.dtype],
+                               Sout=[jax_res.shape],
+                               dim_args_spec=('0.0',))
+
+    res = tf.function(f_tf, jit_compile=True, autograph=False)(x1)
+    self.assertAllClose(
+        tf.nest.map_structure(lambda t: t.numpy(), res), jax_res)
+
+  # TODO(necula): figure out this failure
+  @jtu.skip_on_flag("jax2tf_default_experimental_native_lowering", True)
+  def test_global_device_array(self):
+
+    def create_gda(global_shape, global_mesh, mesh_axes, global_data=None):
+      if global_data is None:
+        global_data = np.arange(np.prod(global_shape)).reshape(global_shape)
+      return GlobalDeviceArray.from_callback(
+          global_shape, global_mesh, mesh_axes,
+          lambda idx: global_data[idx]), global_data
+
+    global_mesh = jtu.create_global_mesh((4, 2), ("x", "y"))
+    mesh_axes = P(("x", "y"))
+    params, _ = create_gda((8, 2), global_mesh, mesh_axes)
+    input_data = np.arange(16).reshape(2, 8)
+
+    # Test 1: use GDA as constants
+    def jax_func(input_data):
+      handle = pjit(
+          jnp.matmul,
+          in_axis_resources=(P("y", "x"), FROM_GDA),
+          out_axis_resources=None)
+      return handle(input_data, params)
+
+    with global_mesh:
+      tf_func = tf.function(
+          jax2tf.convert(jax_func, enable_xla=True),
+          jit_compile=True,
+      )
+      jax_out = jax_func(input_data=input_data)
+      tf_out = tf_func(input_data=input_data)
+      # TODO(b/243146552) We can switch to ConvertAndCompare after this bug fix.
+      np.array_equal(jax_out._value, np.array(tf_out))
+
+    # Test 2: use GDA as JAX function input
+    def jax_func_2(input_data, params):
+      handle = pjit(
+          jnp.matmul,
+          in_axis_resources=(P("y", "x"), P(("x", "y"),)),
+          out_axis_resources=None)
+      return handle(input_data, params)
+
+    with global_mesh:
+      tf_func_2 = tf.function(
+          jax2tf.convert(jax_func_2, enable_xla=True),
+          jit_compile=True,
+      )
+      jax_out_2 = jax_func_2(input_data=input_data, params=params)
+      tf_out_2 = tf_func_2(input_data=input_data, params=params)
+      # TODO(b/243146552) We can switch to ConvertAndCompare after this bug fix.
+      np.array_equal(jax_out_2._value, np.array(tf_out_2))
+
+
 if __name__ == "__main__":
   # TODO: Remove once tensorflow is 2.10.0 everywhere.
-  if not hasattr(tfxla, 'optimization_barrier'):
-    jax.config.update('jax_remat_opt_barrier', False)
+  if not hasattr(tfxla, "optimization_barrier"):
+    jax.config.update("jax_remat_opt_barrier", False)
   absltest.main(testLoader=jtu.JaxTestLoader())
