@@ -1,4 +1,4 @@
-# Copyright 2021 Google LLC
+# Copyright 2021 The JAX Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -23,6 +23,8 @@ from jax.experimental import maps
 from jax.experimental.pjit import pjit, FROM_GDA
 from jax.interpreters.pxla import PartitionSpec as P
 from jax.experimental.global_device_array import GlobalDeviceArray
+from jax._src import distributed
+from jax._src import config as config_internal
 import numpy as np
 
 
@@ -66,9 +68,9 @@ def broadcast_one_to_all(in_tree: PyTreeDef,
   def post_pmap(x):
     return jax.device_get(x)[0]
 
-  in_tree = jax.tree_map(pre_pmap, in_tree)
+  in_tree = jax.tree_util.tree_map(pre_pmap, in_tree)
   in_tree = jax.device_get(_psum(in_tree))
-  return jax.tree_map(post_pmap, in_tree)
+  return jax.tree_util.tree_map(post_pmap, in_tree)
 
 
 def sync_global_devices(name: str):
@@ -101,7 +103,7 @@ def process_allgather(in_tree: PyTreeDef, tiled: bool = False) -> PyTreeDef:
   def _pjit(inp):
     if isinstance(inp, GlobalDeviceArray):
       if inp.is_fully_replicated:
-        return inp.local_data(0).to_py()
+        return np.asarray(inp.local_data(0))
       global_mesh = inp.mesh
       in_axis_resources = FROM_GDA
     else:
@@ -117,16 +119,63 @@ def process_allgather(in_tree: PyTreeDef, tiled: bool = False) -> PyTreeDef:
     with maps.Mesh(global_mesh.devices, global_mesh.axis_names):
       out = pjit(lambda x: x, in_axis_resources=in_axis_resources,
                  out_axis_resources=None)(inp)
-    return out.local_data(0).to_py()
+    return np.asarray(out.local_data(0))
 
-  with jax._src.config.parallel_functions_output_gda(True):
-    return jax.tree_map(_pjit, in_tree)
+  with config_internal.parallel_functions_output_gda(True):
+    return jax.tree_util.tree_map(_pjit, in_tree)
 
 
 def assert_equal(in_tree, fail_message: str = ''):
   """Verifies that all the hosts have the same tree of values."""
   expected = broadcast_one_to_all(in_tree)
   if not jax.tree_util.tree_all(
-      jax.tree_map(lambda *x: np.all(np.equal(*x)), in_tree, expected)):
+      jax.tree_util.tree_map(lambda *x: np.all(np.equal(*x)), in_tree, expected)):
     raise AssertionError(
         f'{fail_message} Expected: {expected}; got: {in_tree}.')
+
+
+def reached_preemption_sync_point(step_id: int) -> bool:
+  """Determine whether all hosts have reached a preemption sync step.
+
+  When any host receive a preemption notice, the notice will be propagated to
+  all hosts and trigger a synchronization protocol in background. The
+  synchronization protocol calculates the maximum step ids from all hosts, and
+  uses the next step id (i.e., max + 1) as the safe step to save a checkpoint.
+  All hosts should continue training more steps until this method returns True,
+  indicating that the `step_id` is equal to the safe step and the hosts should
+  start saving a checkpoint. This feature requires enabling
+  `jax.config.jax_coordination_service`.
+
+  To use this API, all hosts must start training from the same step and call at
+  every training step. Example usage:
+
+  ```
+  def should_save(step_id: int) -> bool:
+
+    # Should save an on-demand checkpoint for preemption
+    if multihost_utils.reached_preemption_sync_point(step_id):
+      return True
+
+    # Should save a regular checkpoint
+    return step_id - last_saved_checkpoint_step >= save_interval_steps
+  ```
+
+  Preemption notice is provided by the cluster scheduler to notify the
+  application in advance before it gets evicted. By default, we use SIGTERM as
+  the signal for preemption notice.
+
+  TODO(b/230630494): Add instructions for customized preemption notice.
+
+  Returns:
+    A boolean indicating whether all hosts have reached a synchronization step
+    after some hosts are preempted.
+
+  Raises:
+    RuntimeError: if preemption sync manager has not been inititialized.
+  """
+  if distributed.global_state.client is None:
+    return False
+  sync_manager = distributed.global_state.preemption_sync_manager
+  if sync_manager is None:
+    raise RuntimeError("Preemption sync manager has not been initialized.")
+  return sync_manager.reached_sync_point(step_id)
