@@ -1,4 +1,4 @@
-# Copyright 2018 Google LLC
+# Copyright 2018 The JAX Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -41,6 +41,7 @@ from jax._src.util import (prod, new_name_stack, safe_zip, safe_map,
 # TODO: update callers to refer to new location.
 from jax._src.util import extend_name_stack as extend_name_stack  # noqa: F401
 from jax._src.util import wrap_name as wrap_name  # noqa: F401
+from jax._src.typing import Shape
 
 from jax._src.lib import xla_bridge as xb
 from jax._src.lib import xla_client as xc
@@ -134,9 +135,9 @@ def parameter(builder, num, shape, name=None, replicated=None):
 # arbitrary tuple nesting, but JAX only uses one level of tupling (and our type
 # checkers don't support recursive types), so we only represent one level of
 # nesting in this type definition.
-SpatialSharding = Union[Tuple[int, ...],
+SpatialSharding = Union[Shape,
                         None,
-                        Tuple[Optional[Tuple[int, ...]], ...]]
+                        Tuple[Optional[Shape], ...]]
 
 def sharding_to_proto(sharding: SpatialSharding):
   """Converts a SpatialSharding to an OpSharding.
@@ -253,12 +254,15 @@ def _canonicalize_python_scalar_dtype(typ, x):
 
 canonicalize_dtype_handlers: Dict[Any, Callable] = {}
 for t in device_array.device_array_types:
-  canonicalize_dtype_handlers[t] = lambda x: x
+  canonicalize_dtype_handlers[t] = identity
 canonicalize_dtype_handlers.update(
     (t, _canonicalize_ndarray_dtype) for t in array_types)
 canonicalize_dtype_handlers.update(
     (t, partial(_canonicalize_python_scalar_dtype, t)) for t in _scalar_types)
-canonicalize_dtype_handlers[core.Token] = lambda x: x
+canonicalize_dtype_handlers[core.Token] = identity
+canonicalize_dtype_handlers[core.PaddedArray] = identity
+canonicalize_dtype_handlers[core.BInt] = \
+    lambda x: core.BInt(_canonicalize_python_scalar_dtype(int, x.val), x.bound)
 
 def abstractify(x) -> core.AbstractValue:
   typ = type(x)
@@ -272,11 +276,16 @@ def abstractify(x) -> core.AbstractValue:
   raise TypeError(f"Argument '{x}' of type '{type(x)}' is not a valid JAX type")
 
 def _make_abstract_python_scalar(typ, val):
-  return ShapedArray((), dtypes._scalar_type_to_dtype(typ, val), weak_type=True)
+  # Note: all python scalar types are weak except bool, because bool only
+  # comes in a single width.
+  return ShapedArray((), dtypes._scalar_type_to_dtype(typ, val),
+                     weak_type=typ is not bool)
 
 pytype_aval_mappings: Dict[Any, Callable[[Any], core.AbstractValue]] = {}
 for t in device_array.device_array_types:
   pytype_aval_mappings[t] = operator.attrgetter('aval')
+pytype_aval_mappings[core.BInt] = lambda x: core.AbstractBInt(x.bound)
+pytype_aval_mappings[core.PaddedArray] = operator.attrgetter('_aval')
 pytype_aval_mappings[core.Token] = lambda _: core.abstract_token
 pytype_aval_mappings.update((t, make_shaped_array) for t in array_types)
 pytype_aval_mappings.update(
